/*			Network File System (NFS) Client
**			   for a WorldWideWeb browser
**			================================
**
**  module written by
**    Ashish Singhal <asinghal@cs.stanford.edu, ashish.singhal@eng.sun.com>
**    Sun Microsystems
**    Mountain View.
*/

#include "nfs_prot3.h"
#include "nfs_prot2.h"
#include "HTNFS.h"
#include "mount.h"
#include <rpc/rpc.h>
#include <stropts.h>
#include <sys/types.h>

#include "HTParse.h"
#include "HTUtils.h"
#include "tcp.h"
#include "HTTCP.h"
#include "HTAnchor.h"
#include "HTFile.h"
#include "HTChunk.h"
#include "HTSort.h"
#include "HText.h"

#ifdef __STDC__
#include <stdlib.h>
#endif

/* #define TRACE 1 */

#ifndef NIL
#define NIL 0
#endif

/*		Hypertext object building machinery
*/
#include "HTML.h"

struct _HTStream {
      CONST HTStreamClass*	isa;
};


/*	Public Declarations
**	--------------------
*/

char*        nfs2ReadLink();
char*        nfs3ReadLink();
BOOL         nfs3StepLookUp();
BOOL         nfs2StepLookUp();
BOOL         nfs3Mount();
BOOL         nfs2Mount();
 
int        errCode  = 0;          /* error code for LookUp and ReadLink */
FILE       *debugFile;
int        NFSDEBUGFLAG = 1;
nfsCache*  nfsCachePtr = (nfsCache*) NULL;
                                  /* ptr to the most recent Hit in the 
                                   * cache                              */
int FirstTime;                    /* checks if a func is being used the first 
                                   * time or not                        */
enum nfsError   nfs_error;


/*	Private Declarations
**	--------------------
*/

PRIVATE LOOKUP3res*  nfs3LookUp();
PRIVATE diropres*    nfs2LookUp();

static int numLinks = 0;          /* number of links encountered        */
static int rootFlag;              /* used for checking STALE and RPCERR */
static int mountPoint;            /* for checking in a dir is a mount 
                                   * point                              */


/*	Filename Comparison Fn
**	----------------------
**
** On entry,
**	first           pointer to a readDir struct
**      second          pointer to a readDir struct
** On exit,
**	returns		-1, 0, 1 depending on the filenames in these two
**			structs
*/
PRIVATE int funcFilenameCmp
ARGS2 (
  const void *,                 first,
  const void *,                 second
)
{
  return strcmp((*(readDir**)first)->filename, (*(readDir**)second)->filename);
}

/*	Read a directory into an hypertext object thru NFSv3
**	----------------------------------------------------
**
** On entry,
**	anchor		Parent anchor to link the this node to
**	address		Address of the directory
**      result_3        Contains the fileid of the Directory
** On exit,
**	returns		HT_LOADED if OK
**			<0 if error.
**
*/
PRIVATE int nfs3ReadDirectory
ARGS6 (
  HTParentAnchor *,		parent,
  CONST char *,			address,
  HTFormat,			format_out,
  HTStream *,			stream,
  nfs_fh3 *,                    Object,
  CLIENT *,                     clnt
  )
{
  HText       *HT = HText_new ();
  HTFormat    format;
  HTAtom      *pencoding;
  char        buffer[BUFSIZE];
  uint32      nEntries = 0;
  extern char *HTgeticonname(HTFormat, char *);
  readDir     *baseReaddirPtr, *ReaddirPtr, *tmpReaddirPtr;
  readDir     **arrayReaddirPtr;
  char *filename = HTParse(address, "", PARSE_PATH + PARSE_PUNCTUATION);
  int         i;

  if(NFSDEBUGFLAG) fprintf(debugFile, "nfs3ReadDir : %s\n", filename);
  
  HText_beginAppend (HT);

  HText_appendText (HT, "<H1>NFS Directory ");
  HText_appendText (HT, filename);
  HText_appendText (HT, "</H1>\n");
  HText_appendText (HT, "<DL>\n");

  /* If this isnt the root level, spit out a parent directory entry */

  if(strcmp(filename,"/") != 0) {
      char    *ptr;

      HText_appendText(HT,"<DD>");

      HText_appendText(HT,"<A HREF=\"");
      
      strcpy(buffer,filename);
      ptr = strrchr(buffer,'/');
      
      if( ptr == &buffer[strlen(buffer) - 1] ) { /* check if filename has */
        if(ptr != NULL) *ptr='\0';		 /* '/' at the end        */
	ptr = strrchr(buffer, '/');
      }
      
      if(ptr != NULL) *ptr='\0';
      
      if(buffer[0] == '\0') 
        HText_appendText(HT,"/");
      else
        HText_appendText(HT, buffer);
      
      HText_appendText(HT,"\"><IMG SRC=\"");
      HText_appendText(HT, HTgeticonname(NULL, "directory"));
      HText_appendText(HT,"\"> Parent Directory</a>");
  }
  
  {

    READDIRPLUS3args  nfsproc3_readdirplus_3_arg;
    READDIRPLUS3res * result_7;
    entryplus3 *      entryPlus3, *tmpEntry;
    char *            msg = "Read %lu entries." ;
    char              line[256];

    memcpy( (void*) &nfsproc3_readdirplus_3_arg.dir, 
	    (void*) Object, 
	    sizeof(nfs_fh3) );
    nfsproc3_readdirplus_3_arg.cookie     = 0;
    strcpy(nfsproc3_readdirplus_3_arg.cookieverf, "0");
    nfsproc3_readdirplus_3_arg.dircount   = (count3) MAX_NFS3_DIRPLUS_INFO;
    nfsproc3_readdirplus_3_arg.maxcount   = (count3) MAX_NFS3_PACKET;
    
    while(TRUE) { 
        int    intr; 
            
        /* update the Mosaic Icon and check for Interrupt */
        intr = HTCheckActiveIcon(1);
        if (intr) {              /* interrupted */
	  HTProgress("Data transfer interrupted");
	  return HT_INTERRUPTED;
        }
   
        result_7 = nfsproc3_readdirplus_3(&nfsproc3_readdirplus_3_arg, clnt);
        if (result_7 == (READDIRPLUS3res *) NULL) {
	    struct rpc_err rpcErr;

	    /* if some entries have been read in, then display them */
	    if(nEntries)
	      break;
	    
	    clnt_geterr(clnt, &rpcErr);
	    if(NFSDEBUGFLAG) clnt_perror(clnt, "call failed");
	    if(FirstTime &&
	       ((rpcErr.re_status == RPC_CANTRECV) )) {
	      nfs_error = NFS_ERR_RPC;
	      return HT_NOT_LOADED;
	    }
	    else
	      return HT_NOT_LOADED;
        }
        if (result_7->status != NFS3_OK) {
	   if(rootFlag && (result_7->status == NFS3ERR_STALE)) {
	     rootFlag = 0;
	     nfs_error = NFS_ERR_STALE;
	     return HT_NOT_LOADED;
 	   }
	   else
             return HT_NOT_LOADED;
	}

        entryPlus3 = result_7->READDIRPLUS3res_u.resok.reply.entries;

        if (entryPlus3 != NULL) {
          /* there are entries to display */ 

	  for( ; entryPlus3 != NULL
	       ; tmpEntry=entryPlus3, entryPlus3 = entryPlus3->nextentry) {

	    if((entryPlus3->name[0] == '.') && (entryPlus3->name[1] == '\0'))
	      continue;

	    if( !strcmp(entryPlus3->name, "..") )
	      continue;

	    sprintf(line, msg, ++nEntries);
	    HTProgress(line);

	    /*  allocate a readDir struct for storing info for sorting 
	     *  later                                                     */
 	    if((tmpReaddirPtr = (readDir*) malloc(sizeof(readDir))) == NULL)
	      continue;
	    tmpReaddirPtr->next = NULL;
	    if((tmpReaddirPtr->filename = strdup(entryPlus3->name))
	       == NULL) {
	      free(tmpReaddirPtr);
	      continue;
	    }
	    tmpReaddirPtr->size     =
		    entryPlus3->name_attributes.post_op_attr_u.attributes.size;
	    tmpReaddirPtr->type     = (int)
		    entryPlus3->name_attributes.post_op_attr_u.attributes.type;
	    if(nEntries == 1) {
	      baseReaddirPtr = ReaddirPtr = tmpReaddirPtr;
	    }
	    else {
	      ReaddirPtr->next = tmpReaddirPtr;
	      ReaddirPtr = tmpReaddirPtr;
	    }
	    
	  } /*  for  */
	} /*  if */

	if(result_7->READDIRPLUS3res_u.resok.reply.eof)
	  break;

	/* update the `cookieverf` val */
	strcpy(nfsproc3_readdirplus_3_arg.cookieverf,
	       result_7->READDIRPLUS3res_u.resok.cookieverf);
	/* update the 'cookie' val */
	nfsproc3_readdirplus_3_arg.cookie = tmpEntry->cookie;
      
    } /* while */
  }

  /* check if this directory is a mount point. if so, get the new 'fh'
   * and read the directory again, else continue on                    */
  if(!nfsCachePtr->PublicFH && mountPoint && !nEntries) {
    nfs_fh3    *tmpObj;
    fattr3     *tmpAttr;
    char       *host = HTParse(address, "", PARSE_HOST);

    mountPoint = 0;
    if( nfs3Mount(host, filename, &tmpObj, &tmpAttr, clnt) 
        && (memcmp( (void*) (Object)->data.data_val,
		    (void*) tmpObj->data.data_val,
		    (Object)->data.data_len )) ) {

      HText_doAbort(HT);
      return nfs3ReadDirectory(parent, address, format_out, stream,
			       tmpObj, clnt);
    }
  }

  /* allocate an array of pointers to sort the filename                */
  arrayReaddirPtr = (readDir **) malloc(sizeof(*arrayReaddirPtr)*nEntries);
  ReaddirPtr = baseReaddirPtr;
  for(i=0; i<nEntries ;i++, ReaddirPtr = ReaddirPtr->next) 
    arrayReaddirPtr[i] = ReaddirPtr;

  qsort((void*)arrayReaddirPtr, nEntries, sizeof(*arrayReaddirPtr),
	funcFilenameCmp);
  
  for(i=0; i<nEntries ;i++)  {
      int   cmpr;

      tmpReaddirPtr = arrayReaddirPtr[i];

      HText_appendText (HT, "\n");
      HText_appendText (HT, "<DD>");
      HText_appendText (HT, "<A HREF=\"");
      /* Assuming it's a relative reference */
      if (tmpReaddirPtr->filename[0] != '/') {
	HText_appendText (HT, filename);
	if (filename[strlen(filename)-1] != '/') 
	  HText_appendText (HT, "/");
      }
      HText_appendText (HT, tmpReaddirPtr->filename);
      HText_appendText (HT, "\"> ");
      
      switch(tmpReaddirPtr->type) {

	    case NF3DIR:
	        HText_appendText(HT, "<IMG SRC=\"");
		HText_appendText(HT, HTgeticonname(NULL, "directory"));
		HText_appendText(HT, "\"> ");
		HText_appendText(HT, tmpReaddirPtr->filename);
      		HText_appendText(HT, "</a>");
	      break;

	    case NF3LNK:
                format = HTFileFormat(tmpReaddirPtr->filename, &pencoding,
		  		      WWW_SOURCE, &cmpr);
	      
		HText_appendText(HT, "<IMG SRC=\"");
		HText_appendText(HT, HTgeticonname(format, "symlink"));
		HText_appendText(HT, "\"> ");
		HText_appendText(HT, tmpReaddirPtr->filename);
      		HText_appendText(HT, "</a>");
              break;

	    case NF3REG:
                format = HTFileFormat(tmpReaddirPtr->filename, &pencoding,
				      WWW_SOURCE, &cmpr);
	      
		HText_appendText(HT, "<IMG SRC=\"");
		/* If we can't figure out what it is,
		   throw up the text icon */
		HText_appendText(HT, HTgeticonname(format, "text")); 

		HText_appendText(HT, "\"> ");
		sprintf(buffer, "%s</a>  (%llu bytes)", tmpReaddirPtr->filename,
		        tmpReaddirPtr->size);
		HText_appendText(HT, buffer);
              break;

	    default:
                HText_appendText(HT, "<IMG SRC=\"");
		HText_appendText(HT, HTgeticonname(NULL, "unknown")); 
		HText_appendText(HT, "\"> ");
		HText_appendText(HT, tmpReaddirPtr->filename);
      		HText_appendText(HT, "</a>");
              break;                  
      } /*  switch */
	    
      free(tmpReaddirPtr->filename);
      free(tmpReaddirPtr);

  } /*  for  */

  HText_appendText(HT, "</DL>\n");
  
  if(NFSDEBUGFLAG) fprintf(debugFile, "      entries: %d\n",nEntries);

  sprintf(buffer, "\n%lu Entries found.", nEntries);
  HText_appendText(HT, buffer);
  HText_endAppend(HT);
  
  return HT_LOADED;
}

PRIVATE int nfs2ReadDirectory
ARGS6 (
  HTParentAnchor *,		parent,
  CONST char *,			address,
  HTFormat,			format_out,
  HTStream *,			stream,
  nfs_fh *,                     Object,
  CLIENT *,                     clnt
  )
{
  HText       *HT = HText_new ();
  HTFormat    format;
  HTAtom      *pencoding;
  char        buffer[BUFSIZE];
  unsigned    nEntries = 0;
  extern char *HTgeticonname(HTFormat, char *);
  readDir     *baseReaddirPtr, *ReaddirPtr, *tmpReaddirPtr;
  readDir     **arrayReaddirPtr;
  char *filename = HTParse(address, "", PARSE_PATH + PARSE_PUNCTUATION);
  int         i;

  if(NFSDEBUGFLAG) fprintf(debugFile, "nfs2ReadDir : %s\n", filename);
  
  HText_beginAppend (HT);

  HText_appendText (HT, "<H1>NFS Directory ");
  HText_appendText (HT, filename);
  HText_appendText (HT, "</H1>\n");
  HText_appendText (HT, "<DL>\n");

  /* If this isnt the root level, spit out a parent directory entry */

  if(strcmp(filename,"/") != 0) {
      char    *ptr;

      HText_appendText(HT,"<DD>");

      HText_appendText(HT,"<A HREF=\"");
      
      strcpy(buffer,filename);
      ptr = strrchr(buffer,'/');
      
      if( ptr == &buffer[strlen(buffer) - 1] ) { /* check if filename has */
        if(ptr != NULL) *ptr='\0';		 /* '/' at the end        */
	ptr = strrchr(buffer, '/');
      }
      
      if(ptr != NULL) *ptr='\0';
      
      if(buffer[0] == '\0') 
        HText_appendText(HT,"/");
      else
        HText_appendText(HT, buffer);
      
      HText_appendText(HT,"\"><IMG SRC=\"");
      HText_appendText(HT, HTgeticonname(NULL, "directory"));
      HText_appendText(HT,"\"> Parent Directory</a>");
  }
  
  {
    readdirargs       nfsproc_readdir_2_arg;
    readdirres *      result_7;
    diropargs         nfsproc_lookup_2_arg;
    entry *           entry2, *tmpEntry;
    char *            msg = "Read %lu entries." ;
    char              line[256];

    memcpy( (void*) &nfsproc_readdir_2_arg.dir, 
	    (void*) Object,
	    sizeof(nfs_fh) );
    memcpy( (void*) &nfsproc_lookup_2_arg.dir,
	    (void*) Object,
	    sizeof(nfs_fh) );
    nfsproc_readdir_2_arg.count   = (unsigned) MAX_NFS2_READ_COUNT;
    memset( (void*) nfsproc_readdir_2_arg.cookie, 0, sizeof(nfscookie) );
    
    while(TRUE) { 
        int    intr; 
            
        /* update the Mosaic Icon and check for Interrupt */
        intr = HTCheckActiveIcon(1);
        if (intr) {              /* interrupted */
	  HTProgress("Data transfer interrupted");
	  return HT_INTERRUPTED;
        }
   
        result_7 = nfsproc_readdir_2(&nfsproc_readdir_2_arg, clnt);
        if (result_7 == (readdirres *) NULL) {
	    struct rpc_err rpcErr;

	    /* if some entries have been read in, then display them */
	    if(nEntries)
	      break;
	    
	    clnt_geterr(clnt, &rpcErr);
	    if(NFSDEBUGFLAG) clnt_perror(clnt, "call failed");
	    if(FirstTime &&
	       ((rpcErr.re_status == RPC_CANTRECV) )) {
	      nfs_error = NFS_ERR_RPC;
	      return HT_NOT_LOADED;
	    }
	    else
	      return HT_NOT_LOADED;
        }
        if (result_7->status != NFS3_OK) {
	  if(rootFlag && (result_7->status == NFSERR_STALE)) {
	     rootFlag = 0;
	     nfs_error = NFS_ERR_STALE;
	     return HT_NOT_LOADED;
 	   }
	   else
             return HT_NOT_LOADED;
	}

        entry2 = result_7->readdirres_u.reply.entries;

        if (entry2 != NULL) {
          diropres  *result_3;
	  /* there are entries to display */ 

	  for( ; entry2 != NULL
	       ; tmpEntry=entry2, entry2 = entry2->nextentry) {
            int   cmpr;

	    if((entry2->name[0] == '.') && (entry2->name[1] == '\0'))
	      continue;

	    if( !strcmp(entry2->name, "..") )
	      continue;

	    /*  get the attributes; do a 'lookup  */
            {
	      if((nfsproc_lookup_2_arg.name = strdup(entry2->name)) == NULL) {
		errCode = errno;
		return NULL;
	      }
 
	      result_3 = nfsproc_lookup_2(&nfsproc_lookup_2_arg, clnt);
	      if (result_3 == (diropres *) NULL) {
		if(NFSDEBUGFLAG)   clnt_perror(clnt, "call failed");
		break;
	      }
	      if (result_3->status != NFS3_OK)
		continue;
	    }

	    sprintf(line, msg, ++nEntries);
	    HTProgress(line);

	    /*  allocate a readDir struct for storing info for sorting 
	     *  later                                                     */
 	    if((tmpReaddirPtr = (readDir*) malloc(sizeof(readDir))) == NULL)
	      continue;
	    tmpReaddirPtr->next = NULL;
	    if((tmpReaddirPtr->filename = strdup(entry2->name))
	       == NULL) {
	      free(tmpReaddirPtr);
	      continue;
	    }
	    tmpReaddirPtr->size =
	                       result_3->diropres_u.diropres.attributes.size;
	    tmpReaddirPtr->type = (int)
	                       result_3->diropres_u.diropres.attributes.type;
	    if(nEntries == 1) {
	      baseReaddirPtr = ReaddirPtr = tmpReaddirPtr;
	    }
	    else {
	      ReaddirPtr->next = tmpReaddirPtr;
	      ReaddirPtr = tmpReaddirPtr;
	    }

	  } /*  for  */
	}   /*  if   */
	    
	if(result_7->readdirres_u.reply.eof)
	  break;

	/* update the `cookie` val */
	memcpy((void*) &nfsproc_readdir_2_arg.cookie,
	       (void*) &tmpEntry->cookie,
	       sizeof(nfscookie) );
      
    } /* while */
  }
  
  /* check if this directory is a mount point. if so, get the new 'fh'
   * and read the directory again, else continue on                    */
  if(!nfsCachePtr->PublicFH && mountPoint && !nEntries) {
    nfs_fh     *tmpObj;
    fattr      *tmpAttr;
    char       *host = HTParse(address, "", PARSE_HOST);

    mountPoint = 0;
    if( nfs2Mount(host, filename, &tmpObj, &tmpAttr, clnt) 
        && memcmp((void*) Object, (void*) tmpObj, sizeof(nfs_fh)) ) {
      
      HText_doAbort(HT);
      return nfs2ReadDirectory(parent, address, format_out, stream,
			       tmpObj, clnt);
    }
  }
  
  /* allocate an array of pointers to sort the filename                */
  arrayReaddirPtr = (readDir **) malloc(sizeof(*arrayReaddirPtr)*nEntries);
  ReaddirPtr = baseReaddirPtr;
  for(i=0; i<nEntries ;i++, ReaddirPtr = ReaddirPtr->next) 
    arrayReaddirPtr[i] = ReaddirPtr;

  qsort((void*)arrayReaddirPtr, nEntries, sizeof(*arrayReaddirPtr),
	funcFilenameCmp);
  
  for(i=0; i<nEntries ;i++)  {
      int   cmpr;

      tmpReaddirPtr = arrayReaddirPtr[i];

      HText_appendText (HT, "\n");
      HText_appendText (HT, "<DD>");
      HText_appendText (HT, "<A HREF=\"");
      /* Assuming it's a relative reference */
      if (tmpReaddirPtr->filename[0] != '/') {
	  HText_appendText (HT, filename);
	  if (filename[strlen(filename)-1] != '/') 
	    HText_appendText (HT, "/");
      }
      HText_appendText (HT, tmpReaddirPtr->filename);
      HText_appendText (HT, "\"> ");

      switch(tmpReaddirPtr->type) {

	 case NF3DIR:
	        HText_appendText(HT, "<IMG SRC=\"");
		HText_appendText(HT, HTgeticonname(NULL, "directory"));
		HText_appendText(HT, "\"> ");
		HText_appendText(HT, tmpReaddirPtr->filename);
      		HText_appendText(HT, "</a>");
	      break;

	 case NF3LNK:
                format = HTFileFormat(tmpReaddirPtr->filename, &pencoding,
		  		      WWW_SOURCE, &cmpr);
	      
		HText_appendText(HT, "<IMG SRC=\"");
		HText_appendText(HT, HTgeticonname(format, "symlink"));
		HText_appendText(HT, "\"> ");
		HText_appendText(HT, tmpReaddirPtr->filename);
      		HText_appendText(HT, "</a>");
              break;

	 case NF3REG:
                format = HTFileFormat(tmpReaddirPtr->filename, &pencoding,
				      WWW_SOURCE, &cmpr);
	      
		HText_appendText(HT, "<IMG SRC=\"");
		/* If we can't figure out what it is,
		   throw up the text icon */
		HText_appendText(HT, HTgeticonname(format, "text")); 

		HText_appendText(HT, "\"> ");
		sprintf(buffer, "%s</a>  (%u bytes)", tmpReaddirPtr->filename,
		        (unsigned)tmpReaddirPtr->size);
		HText_appendText(HT, buffer);
              break;

	 default:
                HText_appendText(HT, "<IMG SRC=\"");
		HText_appendText(HT, HTgeticonname(NULL, "unknown")); 
		HText_appendText(HT, "\"> ");
		HText_appendText(HT, tmpReaddirPtr->filename);
      		HText_appendText(HT, "</a>");
              break;                  
      } /*  switch */
	    
      free(tmpReaddirPtr->filename);
      free(tmpReaddirPtr);

  } /*  for  */

  HText_appendText(HT, "</DL>\n");

  if(NFSDEBUGFLAG) fprintf(debugFile, "      entries: %d\n",nEntries);
  sprintf(buffer, "\n%lu Entries found.", nEntries);
  HText_appendText(HT, buffer);
  HText_endAppend(HT);
  
  return HT_LOADED;
}

/*	Does a NFS read of the File and puts it in the Mosaic Window
**	------------------------------------------------------------
**
** On entry,
**	filename	name of the file
** On exit,
**	returns		0 if success 
**			<0 if failure
*/
PRIVATE int nfs3ReadFile
ARGS7 (
  char *,			filename,
  HTParentAnchor *,		anchor,
  HTFormat,			format_out,
  HTStream *,			sink,
  nfs_fh3 *,                    Object,
  fattr3 *,                     Attributes,
  CLIENT *,			clnt
)
{
  HTStream *     stream;
  HTStreamClass  targetClass;    
  HTAtom *       encoding;
  HTFormat       format;
  int            compressed = 0;
  uint64         fileSize = Attributes->size;

  if(NFSDEBUGFLAG) fprintf(debugFile, "nfs3ReadFile : %s (%llu)\n",
			   filename, fileSize);
  
  /* append '/' if not there */
  if (! (int)(*filename))
     StrAllocCopy(filename, "/");
  format = HTFileFormat(filename, &encoding, WWW_PLAINTEXT, &compressed);
              
  stream = HTStreamStack(format, format_out, compressed, sink, anchor);
      
  if (!stream) 
  {
    char buffer[1024];	/* @@@@@@@@ */
    sprintf(buffer, "Sorry, can't convert from %s to %s.",
                  HTAtom_name(format), HTAtom_name(format_out));
    HTProgress (buffer);
    if(NFSDEBUGFLAG) fprintf(debugFile, "NFS: %s\n", buffer);
    return HT_NOT_LOADED;
  }
      
  targetClass = *(stream->isa);	/* Copy pointers to procedures */

  /* read in the file */
  {
    READ3args        nfsproc3_read_3_arg;
    READ3res         *result_5;
    uint64           nBytes=0;
    char             *msg = "Read %llu of %llu bytes of data." ;
    char             line[256];
    extern int       twirl_increment;
    int              next_twirl = twirl_increment;
    count3           nMaxReadCount = (count3) MAX_NFS3_READ_COUNT;


    memcpy( (void*) &nfsproc3_read_3_arg.file, 
	    (void*) Object, 
	    sizeof(nfs_fh3) );
    nfsproc3_read_3_arg.count = nMaxReadCount;

    
    while(TRUE) {
        int     intr;

	/* update the Mosaic Icon and check for Interrupt */
	if (nBytes > next_twirl) {
	  intr = HTCheckActiveIcon(1);
	  next_twirl += twirl_increment;
	}
	else {
	  intr = HTCheckActiveIcon(0);
	}

	if (intr) {              /* interrupted */
	  HTProgress("Data transfer interrupted");
	  (*targetClass.handle_interrupt)(stream);
	  return HT_INTERRUPTED;
	}

	nfsproc3_read_3_arg.offset = nBytes;

	result_5 = nfsproc3_read_3(&nfsproc3_read_3_arg, clnt);
	if (result_5 == (READ3res *) NULL) {
	  struct rpc_err rpcErr;
	  clnt_geterr(clnt, &rpcErr);
	  if(NFSDEBUGFLAG) clnt_perror(clnt, "call failed");
	  if(FirstTime &&
	     ((rpcErr.re_status == RPC_CANTRECV) )) {
	    nfs_error = NFS_ERR_RPC;
	    return HT_NOT_LOADED;
	  }
	  else
	    return HT_NOT_LOADED;
	}
	if (result_5->status != NFS3_OK) 
	  return HT_NOT_LOADED;

	/* put the info on the Mosaic window */
	(*targetClass.put_block)(stream,
				 result_5->READ3res_u.resok.data.data_val, 
				 result_5->READ3res_u.resok.count);

	/* check if the File has been completely read */
	if (result_5->READ3res_u.resok.eof) {
	  HTProgress("Data transfer complete");
	  break;
	}

	if(result_5->READ3res_u.resok.count < nMaxReadCount)
          nfsproc3_read_3_arg.count = nMaxReadCount
		                    = result_5->READ3res_u.resok.count;
	  
	/* print Progress message */
	sprintf(line, msg, nBytes, fileSize);
	HTProgress(line);

	nBytes += (uint64) result_5->READ3res_u.resok.count ;
	
    } /* while */
  }
  
  (*targetClass.end_document)(stream);
  (*targetClass.free)(stream);
      
  return HT_LOADED;
}       

/*	Does a NFS read of the File and puts it in the Mosaic Window
**	------------------------------------------------------------
**
** On entry,
**	filename	name of the file
** On exit,
**	returns		0 if success 
**			<0 if failure
*/
PRIVATE int nfs2ReadFile
ARGS7 (
  char *,			filename,
  HTParentAnchor *,		anchor,
  HTFormat,			format_out,
  HTStream *,			sink,
  nfs_fh *,                     Object,
  fattr *,                      Attributes,
  CLIENT *,			clnt
)
{
  HTStream *     stream;
  HTStreamClass  targetClass;    
  HTAtom *       encoding;
  HTFormat       format;
  int            compressed = 0;
  unsigned       fileSize = Attributes->size;

  if(NFSDEBUGFLAG) fprintf(debugFile, "nfs2ReadFile : %s (%d)\n",
			    filename, fileSize);
  
  /* append '/' if not there */
  if (! (int)(*filename))
     StrAllocCopy(filename, "/");
  format = HTFileFormat(filename, &encoding, WWW_PLAINTEXT, &compressed);
              
  stream = HTStreamStack(format, format_out, compressed, sink, anchor);
      
  if (!stream) 
  {
    char buffer[1024];	/* @@@@@@@@ */
    sprintf(buffer, "Sorry, can't convert from %s to %s.",
                  HTAtom_name(format), HTAtom_name(format_out));
    HTProgress (buffer);
    if(NFSDEBUGFLAG) fprintf(debugFile, "NFS: %s\n", buffer);
    return HT_NOT_LOADED;
  }
      
  targetClass = *(stream->isa);	/* Copy pointers to procedures */

  /* read in the file */
  {
    readargs         nfsproc_read_2_arg;
    readres          *result_5;
    unsigned         nBytes=0;
    char             *msg = "Read %u of %u bytes of data." ;
    char             line[256];
    extern int       twirl_increment;
    int              next_twirl = twirl_increment;


    memcpy( (void*) &nfsproc_read_2_arg.file, 
	    (void*) Object,
	    sizeof(nfs_fh) );
    nfsproc_read_2_arg.count = (unsigned) MAX_NFS2_READ_COUNT;

    
    while(TRUE) {
        int     intr;

	/* update the Mosaic Icon and check for Interrupt */
	if (nBytes > next_twirl) {
	  intr = HTCheckActiveIcon(1);
	  next_twirl += twirl_increment;
	}
	else {
	  intr = HTCheckActiveIcon(0);
	}

	if (intr) {              /* interrupted */
	  HTProgress("Data transfer interrupted");
	  (*targetClass.handle_interrupt)(stream);
	  return HT_INTERRUPTED;
	}

	nfsproc_read_2_arg.offset = nBytes;

	result_5 = nfsproc_read_2(&nfsproc_read_2_arg, clnt);
	if (result_5 == (readres *) NULL) {
	  struct rpc_err rpcErr;
	  clnt_geterr(clnt, &rpcErr);
	  if(NFSDEBUGFLAG) clnt_perror(clnt, "call failed");
	  if(FirstTime &&
	     ((rpcErr.re_status == RPC_CANTRECV) )) {
	    nfs_error = NFS_ERR_RPC;
	    return HT_NOT_LOADED;
	  }
	  else
	    return HT_NOT_LOADED;
	}
	if (result_5->status != NFS3_OK) 
	  return HT_NOT_LOADED;

	/* put the info on the Mosaic window */
	(*targetClass.put_block)(stream,
				 result_5->readres_u.reply.data.data_val, 
				 result_5->readres_u.reply.data.data_len);

	/* check if the File has been completely read */
	if( (fileSize -= result_5->readres_u.reply.data.data_len) == 0) {
	  HTProgress("Data transfer complete");
	  break;
	}

	/* print Progress message */
	sprintf(line, msg, nBytes, fileSize);
	HTProgress(line);

	nBytes += (unsigned) result_5->readres_u.reply.data.data_len ;
	
    } /* while */
  }
  
  (*targetClass.end_document)(stream);
  (*targetClass.free)(stream);
      
  return HT_LOADED;
}       

/*      Read a Symbolic Link 
**	--------------------
**
** On entry,
**	symlink_fh      file handle of the symbolic link
**      dir_fh          directory handle of the dir in which the link exists
** On exit,
**	nfs_fh3         file handle of the symbolic link
**                      NULL if error
*/

char* nfs3ReadLink
ARGS3 (
  char *,          hostname,
  nfs_fh3 *,	   symlink_fh,
  CLIENT *,        clnt
)
{
      READLINK3args  nfsproc3_readlink_3_arg;
      READLINK3res   *result_4;
      LOOKUP3res     *result_3;

      if(numLinks++ > MAXSYMLINKS) {
	if(NFSDEBUGFLAG) fprintf(debugFile, ".... Possible Loop on Links\n");
	errCode = HT_NOT_LOADED;
	return NULL;
      }

      /*  check for interruption  */
      if(HTCheckActiveIcon(0)) {
	HTProgress("Data transfer interrupted.");
	errCode = HT_INTERRUPTED;
	return NULL;
      }
      
      memcpy( (void*) &nfsproc3_readlink_3_arg.symlink,
	      (void*) symlink_fh,
	      sizeof(nfs_fh3) );
	
      result_4 = nfsproc3_readlink_3(&nfsproc3_readlink_3_arg, clnt);
      if (result_4 == (READLINK3res *) NULL) {
	errCode =  HT_NOT_LOADED;
	return NULL;
      }
      if (result_4->status != NFS3_OK) {
	errCode = HT_NOT_LOADED;
        return NULL;
      }

      return result_4->READLINK3res_u.resok.data;
}


/*      Read a Symbolic Link (ver 2)
**	--------------------
**
** On entry,
**	symlink_fh      file handle of the symbolic link
**      dir_fh          directory handle of the dir in which the link exists
** On exit,
**	nfs_fh3         file handle of the symbolic link
**                      NULL if error
*/

char* nfs2ReadLink
ARGS3 (
  char *,          hostname,
  nfs_fh *,	   symlink_fh,
  CLIENT *,        clnt
)
{
      nfs_fh         nfsproc_readlink_2_arg;
      readlinkres    *result_4;
      diropres       *result_3;

      if(numLinks++ > MAXSYMLINKS) {
	errCode = HT_NOT_LOADED;
	return NULL;
      }

      /*  check for interruption  */
      if(HTCheckActiveIcon(0)) {
	HTProgress("Data transfer interrupted.");
	errCode = HT_INTERRUPTED;
	return NULL;
      }
      
      memcpy( (void*) &nfsproc_readlink_2_arg,
	      (void*) symlink_fh,
	      sizeof(nfs_fh) );
	
      result_4 = nfsproc_readlink_2(&nfsproc_readlink_2_arg, clnt);
      if (result_4 == (readlinkres *) NULL) {
	if(NFSDEBUGFLAG) clnt_perror(clnt, "call failed");
	errCode = HT_NOT_LOADED;
	return NULL;
      }
      if (result_4->status != NFS3_OK) {
	errCode = HT_NOT_LOADED;
        return NULL;
      }

      return result_4->readlinkres_u.data;
}


/*      LookUp thru the directories 
**	---------------------------
**
** On entry,
**	filename	address of a file/dir
** On exit,
**	result_3        result of lookup of the file/dir
**                      NULL if error
*/

PRIVATE LOOKUP3res* nfs3LookUp
ARGS3 (
  char *,          hostname,
  char *,	   filename,
  CLIENT *,        clnt
)
{
    LOOKUP3res  *result_3;
    LOOKUP3args  nfsproc3_lookup_3_arg;
    
  if(NFSDEBUGFLAG) fprintf(debugFile, "nfs3Lookup : %s:%s\n",
			    hostname, filename);
  
    /*  check for interruption  */
    if(HTCheckActiveIcon(0)) {
      HTProgress("Data transfer interrupted.");
      errCode = HT_INTERRUPTED;
      return NULL;
    }

    if ( filename[0] == '/' ) 
      memcpy( (void*) filename, (void*) &filename[1], strlen(filename) );
    if ( filename[strlen(filename)-1] == '/' )
      filename[strlen(filename)-1] = '\0';

    /*  get the file handle  */
    {
      memset( (void*) &nfsproc3_lookup_3_arg.what.dir, 0, sizeof(nfs_fh3) );

      if ( (nfsproc3_lookup_3_arg.what.name = strdup(filename)) == NULL) {
	errCode = errno;
	return NULL;
      }

      result_3 = nfsproc3_lookup_3(&nfsproc3_lookup_3_arg, clnt);
      if (result_3 == (LOOKUP3res *) NULL) {
	  struct rpc_err rpcErr;
	  
	  free(nfsproc3_lookup_3_arg.what.name);
	  clnt_geterr(clnt, &rpcErr);

	  if(NFSDEBUGFLAG) clnt_perror(clnt, "call failed");
	  if(FirstTime &&
	     ((rpcErr.re_status == RPC_CANTRECV) ))
	    nfs_error = NFS_ERR_RPC;
	  else
	    errCode =  HT_NOT_LOADED;
	  return NULL;
      }

      if (result_3->status != NFS3_OK) {
    	free(nfsproc3_lookup_3_arg.what.name);
	if(result_3->status == NFS3ERR_STALE)
	  nfs_error = NFS_ERR_STALE;
	else
	  errCode = HT_NOT_LOADED;
        return NULL;
      }

      /*  check if the 'fh' is a link  */
      if(result_3->LOOKUP3res_u.resok.obj_attributes.post_op_attr_u.attributes.type == NF3LNK) {

	 char *sym_lnk_data, *dirname = strrchr(filename, '/');

	 /*   read the link  */
	 if( (sym_lnk_data = nfs3ReadLink(hostname,
					  &result_3->LOOKUP3res_u.resok.object,
					  clnt))
	     == NULL) 
	   return NULL;

	 /*  if the symbolic link is wrt to the root dir, then the URL is 
	  *  redirected to that using the 'file' WWW protocol             */
	 if(sym_lnk_data[0] == '/') {
 	   extern char *redirecting_url;

	   StrAllocCopy(redirecting_url, "file://");
	   StrAllocCat (redirecting_url, hostname);
	   StrAllocCat (redirecting_url, sym_lnk_data);
	   errCode = HT_REDIRECTING;
	   return NULL;
	 }

	/*
	 * otherwise, if it looks like a URL then
	 * pass it up for re-evaluation.
	 */
	else if(strstr(sym_lnk_data,"://") != 0) {
	   extern char *redirecting_url; 
	   StrAllocCopy (redirecting_url, sym_lnk_data);
	   errCode = HT_REDIRECTING;
	   return NULL;
	 }


	 if(dirname != NULL) {
	   *(dirname+1) = '\0';
	   StrAllocCat(filename, sym_lnk_data);
	 }
	 else
	   StrAllocCopy(filename, sym_lnk_data);

	 return nfs3LookUp(hostname, filename, clnt);
      }
    }
      
    free(nfsproc3_lookup_3_arg.what.name);
 
    return result_3;
}


/*      LookUp thru the directories 
**	---------------------------
**
** On entry,
**	filename	address of a file/dir
** On exit,
**	result_3        result of lookup of the file/dir
**                      NULL if error
*/

PRIVATE diropres* nfs2LookUp
ARGS3 (
  char *,          hostname,
  char *,	   filename,
  CLIENT *,        clnt
)
{
    diropres     *result_3;
    diropargs    nfsproc_lookup_2_arg;

  if(NFSDEBUGFLAG) fprintf(debugFile, "nfs2Lookup : %s:%s\n",
			    hostname, filename);

    /*  check for interruption  */
    if(HTCheckActiveIcon(0)) {
      HTProgress("Data transfer interrupted.");
      errCode = HT_INTERRUPTED;
      return NULL;
    }

    if ( filename[0] == '/' ) 
      memcpy( (void*) filename, (void*) &filename[1], strlen(filename) );
    if ( filename[strlen(filename)-1] == '/' )
      filename[strlen(filename)-1] = '\0';

    /*  get the file handle  */
    {
      memset( (void*) &nfsproc_lookup_2_arg.dir, 0, sizeof(nfs_fh) );

      if ( (nfsproc_lookup_2_arg.name = strdup(filename)) == NULL) {
	errCode = errno;
        return NULL;
      }
      
      result_3 = nfsproc_lookup_2(&nfsproc_lookup_2_arg, clnt);
      if (result_3 == (diropres *) NULL) {
	  struct rpc_err rpcErr;
	  
	  free(nfsproc_lookup_2_arg.name);
	  clnt_geterr(clnt, &rpcErr);

	  if(NFSDEBUGFLAG) clnt_perror(clnt, "call failed");
	  if(FirstTime &&
	     ((rpcErr.re_status == RPC_CANTRECV) ))
	    nfs_error =  NFS_ERR_RPC;
	  else
	    errCode =  HT_NOT_LOADED;
	  return NULL;
      }

      free(nfsproc_lookup_2_arg.name);

      if (result_3->status != NFS3_OK) {
	if(result_3->status == NFSERR_STALE)
	  nfs_error = NFS_ERR_STALE;
	else
	  errCode = HT_NOT_LOADED;
        return NULL;
      }
    
      /* check if the 'fh' is a link */
      if(result_3->diropres_u.diropres.attributes.type == NFLNK) {

	 char     *sym_lnk_data, *dirname = strrchr(filename, '/');
	 
	 if((sym_lnk_data = nfs2ReadLink(hostname,
				&result_3->diropres_u.diropres.file,
				clnt))
	    == NULL)
	    return NULL;

	 /*  if the symbolic link is wrt to the root dir, then the URL is 
	  *  redirected to that using the 'file' WWW protocol             */
	 if(sym_lnk_data[0] == '/') {
 	   extern char *redirecting_url;

	   StrAllocCopy(redirecting_url, "file://");
	   StrAllocCat (redirecting_url, hostname);
	   StrAllocCat (redirecting_url, sym_lnk_data);
	   errCode = HT_REDIRECTING;
	   return NULL;
	 }

	 if(dirname != NULL) {
	   *(dirname+1) = '\0';
	   StrAllocCat(filename, sym_lnk_data);
	 }
	 else
	   StrAllocCopy(filename, sym_lnk_data);

	 return nfs2LookUp(hostname, filename, clnt);
       }
    }
 
    return result_3;
}

/*      Create a New Client 
**	-------------------
**
** On entry,
**	hostname	name of the host
**      CachePtr        ptr to a 'nfsCache' entry
** On exit,
**	returns		1 if successful 
*/
PRIVATE int CreateNewClient
ARGS3 (
  char *,                      hostname,
  char *,                      url,
  nfsCache **,                 NFSCachePtr
)
{
   nfsCache *     CachePtr = *NFSCachePtr;
   char *         strport = NULL;
   int            con = -1, status = 0;

   if(NFSDEBUGFLAG) fprintf(debugFile, "    CreateNewClient : %s\n", url);


   /* Parse port number if present */
   if (strport=strchr(hostname, ':')) {
      *strport++ = 0;              /* Chop off port */
      if (strport[0]>='0' && strport[0]<='9'
	  && (atol(strport) != IPPORT_NFS)) {  /* if port = 2049, fall back 
                                                * to the normal routine */
	/* try to create a client at the 'port'     */
	if(CreatePortClient(hostname, atol(strport),
			    NFS_PROGRAM, NFSCachePtr))
	   goto return_ok;
	else
	   goto return_error;
      }
   }    

   /*  go thru the various options          */
   if( ((CachePtr->clnt
	    = clnt_create(hostname, NFS_PROGRAM, NFS_V3, "tcp"))
        != (CLIENT*) NULL)

      || (HTCheckActiveIcon(0))      /*  check for interruption  */
      || ((CachePtr->clnt
	       = clnt_create(hostname, NFS_PROGRAM, NFS_V3, "udp"))
	  != (CLIENT*) NULL) )
   {
	char* nfsproc3_null_3_arg;
	void  *result_1;

	if(CachePtr->clnt == NULL) {
          HTProgress("Data transfer interrupted.");
          return 0;
        }
		   
        /*  do a NULL rpc and check for 'valid' clnt handle  */
	result_1= nfsproc3_null_3((void *)&nfsproc3_null_3_arg, CachePtr->clnt);
	if (result_1 != (void *) NULL) {
	  CachePtr->Version = NFS_V3;
	  goto return_ok;
	}
   }
   
   /*  check for interruption  */
   if(HTCheckActiveIcon(0)) {
      HTProgress("Data transfer interrupted.");
      goto return_error;
   }
   
   if( (CachePtr->clnt
	 = clnt_create(hostname, NFS_PROGRAM, NFS_VERSION, "tcp")) 
      != (CLIENT*) NULL)
      CachePtr->Version = NFS_VERSION;
   else if(HTCheckActiveIcon(0)) {     /*  check for interruption  */
       HTProgress("Data transfer interrupted.");
       return 0;
   }
   else if( (CachePtr->clnt
	 = clnt_create(hostname, NFS_PROGRAM, NFS_VERSION, "udp"))
      != (CLIENT*) NULL)
      CachePtr->Version = NFS_VERSION;
   else if(HTCheckActiveIcon(0)) {     /*  check for interruption  */
       HTProgress("Data transfer interrupted.");
       goto return_error;
   }
   else
     /* proxy connect */
    if((status = HTDoConnect(url, "NFS", IPPORT_NFS, &CachePtr->fd)) >= 0){
      struct netconfig *nc = getnetconfigent("tcp");
      
      if( (ioctl(con, I_POP, "sockmod") != 0) )
	if(NFSDEBUGFLAG) fprintf(debugFile, "%d", errno);
      if (ioctl(con, I_PUSH, "timod") != 0)
	if(NFSDEBUGFLAG) fprintf(debugFile, "%d", errno);

      if ( (CachePtr->clnt
                = clnt_tli_create(con, nc, NULL, NFS_PROGRAM, NFS_V3, 0, 0))
              != (CLIENT*) NULL)
	 CachePtr->Version = NFS_V3;

      else if(HTCheckActiveIcon(0)) {   /*  check for interruption  */
	HTProgress("Data transfer interrupted.");
	goto return_error;
      }

      else if ((CachePtr->clnt
               = clnt_tli_create(con, nc, NULL, NFS_PROGRAM, NFS_VERSION, 0, 0))
              != (CLIENT*) NULL)  
	 CachePtr->Version = NFS_VERSION;
      else {
          if (NFSDEBUGFLAG) {
            clnt_pcreateerror(hostname);
            fprintf (debugFile, "HTNFSLoad: Unable to Create Client");
          }
	  goto return_error;
      }
    }
    else {

      if(status = HT_INTERRUPTED) {
        HTProgress ("Connection interrupted.");
	if(con != -1) {
	  NETCLOSE(con);
	  con = -1;
        }
        goto return_error;
      }
 
      if (NFSDEBUGFLAG) {
        clnt_pcreateerror(hostname);
        fprintf (stderr, "HTNFSLoad: Unable to Create Client");
      }
      goto return_error;
    }

return_ok:	
   CachePtr->clnt->cl_auth = authsys_create_default();

   return 1;

return_error:
  if(CachePtr->fd != -1)
    NETCLOSE(CachePtr->fd);
  CachePtr->fd = -1;
  free(CachePtr->hostname);
  return 0;
}


/*      Replace an Old Client with a New One
**	------------------------------------
**
** On entry,
**	hostname	name of the host
**      CachePtr        ptr to a 'nfsCache' entry
** On exit,
**	returns		1 if successful 
*/

PRIVATE BOOL ReplaceClient
ARGS4(
  char *,                      hostname,
  char *,                      url,
  CLIENT **,                   clnt,
  int **,                      Ver
)
{
  if(NFSDEBUGFLAG) fprintf(debugFile, "Replace Client: %s by %s\n",
			    hostname, nfsCachePtr->hostname);

  /* destroy old client  */
  clnt_destroy(nfsCachePtr->clnt);
  if(nfsCachePtr->fd != -1)
    t_close(nfsCachePtr->fd);
  if(nfsCachePtr->tbind)
    t_free((char *) nfsCachePtr->tbind, T_BIND);
  
  if(nfsCachePtr->prefixtable != NULL) {
    prefix  *tmp, *pre = nfsCachePtr->prefixtable;
    
    for(; pre != NULL ; pre = tmp) {
      free(pre->pathname);
      if(nfsCachePtr->Version == NFS_V3) {
	  free( ((nfs_fh3*)pre->fh)->data.data_val);
	  free(  (nfs_fh3*)pre->fh);
      }
      else
	  free((nfs_fh*)pre->fh);
      tmp = pre->next;
      free(pre);
    }
    
    nfsCachePtr->prefixtable = NULL;
  }
  
  nfsCachePtr->fd    = -1;
  nfsCachePtr->tbind = NULL;
  
  if(!CreateNewClient(hostname, url, &nfsCachePtr)) {
    nfsCache *    tmpPtr = nfsCachePtr;
    
    nfsCachePtr->prev->next = nfsCachePtr->next;
    nfsCachePtr->next->prev = nfsCachePtr->prev;
    nfsCachePtr = nfsCachePtr->prev;
    
    free(tmpPtr->hostname);
    free(tmpPtr);
    return 0;
  }

  *clnt = nfsCachePtr->clnt;
  **Ver = nfsCachePtr->Version;
  return 1;
}


/*      Search the NFS Cache for Active Connections
**	-------------------------------------------
**
** On entry,
**	hostname	name of the host
** On exit,
**	returns		1 if successful in either a hit in the cache
**                                       or a insertion in the cache
**			0 if not.
** Cache Implementation:
**     The Cache is a circular linked list having :-
**              'hostname' - name of the host,
**              'clnt'     - connection struct,
**              'MountClnt'- handle for mount proto (when needed)
**              'fd'       - file descriptor 
**                                * used for connection to proxy servers
**                                * and  for connection to nfs servers on
**                                    arbitrary ports
**              'tbind'    - used for connection to nfs servers on arbitrary
**                             ports 
**              'next'     - next cache element,
**              'prev'     - prev cache element
**    
**     A pointer 'nfsCachePtr' is maintained which points to the most
**     recent Hit in the Cache.
**     A New cache element is inserted after 'nfsCachePtr'.
**     The searching is done from the nfsCachePtr going thru the 'prev'
**     elements.
*/
PRIVATE BOOL SearchNfsCache
ARGS4(
  char *,                      hostname,
  char *,                      url,
  CLIENT **,                   clnt,
  int *,                       Ver
)
{
  nfsCache *        tmpPtr;
  static int        nEntries = 0;
  int               con = -1, status = 0;

  if(NFSDEBUGFLAG) fprintf(debugFile, "  SearchNFSCache : %s\n", url);

  /*  check if we have not yet have cached anything yet   */
  if(nfsCachePtr == (nfsCache*) NULL) {
    if( (nfsCachePtr = (nfsCache*) malloc(sizeof(nfsCache))) == NULL)
       return 0;
    if( (nfsCachePtr->hostname = strdup(hostname)) == NULL) {
       free(nfsCachePtr);
       nfsCachePtr = (nfsCache*) NULL;
       return 0;
    }
    /* init */
    nfsCachePtr->fd    = -1;
    nfsCachePtr->tbind = NULL;
    
    /*  create a new client; try out the various options  */
    if(!CreateNewClient(hostname, url, &nfsCachePtr)) {
      free(nfsCachePtr);
      nfsCachePtr = (nfsCache*) NULL;
      return 0;
    }

    nfsCachePtr->PublicFH = TRUE;
    nfsCachePtr->prefixtable = NULL;

    *clnt = nfsCachePtr->clnt;
    *Ver  = nfsCachePtr->Version;
    nEntries++;
    nfsCachePtr->next = nfsCachePtr->prev = nfsCachePtr;
    return 1;
  }

  /*  search thru the cache */
  tmpPtr = nfsCachePtr;
  do {
    if( !strcmp(tmpPtr->hostname, hostname) ) {        /*  Hit!  */
	*clnt = tmpPtr->clnt ;
        *Ver  = tmpPtr->Version;
	nfsCachePtr = tmpPtr;
        return 1;
    }

    tmpPtr = tmpPtr->prev;
  } while(tmpPtr != nfsCachePtr);

  
  /*  create a new struct to put the info into  */
  if(nEntries != NFSCLIENTCACHESIZE) {
    /*  insert the struct bet `nfsCachePtr` and its `next one`  */
    if( (tmpPtr = (nfsCache*) malloc(sizeof(nfsCache))) == NULL) 
       return 0;
    tmpPtr->next            = nfsCachePtr->next;
    nfsCachePtr->next->prev = tmpPtr;
    nfsCachePtr->next       = tmpPtr;
    tmpPtr->prev            = nfsCachePtr;
    tmpPtr->fd    = -1;
    tmpPtr->tbind = NULL;
   }
  else {
    /*  remove the `next struct`                */
    tmpPtr = nfsCachePtr->next;
    free(tmpPtr->hostname);
    /* destroy client */
    clnt_destroy(tmpPtr->clnt);
    if(tmpPtr->fd != -1)
      t_close(tmpPtr->fd);
    if(tmpPtr->tbind)
      t_free((char *) tmpPtr->tbind, T_BIND);
    
    if(tmpPtr->prefixtable != NULL) {
      prefix  *tmp, *pre = tmpPtr->prefixtable;
      for(; pre != NULL ; pre = tmp) {
	free(pre->pathname);
	if(tmpPtr->Version == NFS_V3) {
	  free( ((nfs_fh3*)pre->fh)->data.data_val);
	  free(  (nfs_fh3*)pre->fh);
	}
	else
	  free((nfs_fh*)pre->fh);
	tmp = pre->next;
	free(pre);
      }
    
      tmpPtr->prefixtable = NULL;
    }
      
    nEntries--;
    /* yup, in case of error below, the above info is lost ... */
  }
  
  /*  duplicate the name     */
  if( (tmpPtr->hostname = strdup(hostname)) == NULL) {
    nfsCachePtr->next = tmpPtr->next;
    tmpPtr->next->prev = nfsCachePtr;
    free(tmpPtr);
    return 0;
  }

  /*  create a new client    */
  if(!CreateNewClient(hostname, url, &tmpPtr)) {
      nfsCachePtr->next = tmpPtr->next;
      tmpPtr->next->prev = nfsCachePtr;
      free(tmpPtr);
      return 0;
  }
  
  tmpPtr->PublicFH = TRUE;
  tmpPtr->prefixtable = NULL;
  nEntries++;
  nfsCachePtr = tmpPtr;
  *clnt = nfsCachePtr->clnt;
  *Ver  = nfsCachePtr->Version;
  
  return 1;
}

/*	Perform NFS v3 processing
**	-------------------------
**
** On entry,
**	name		WWW address of a file: document, including hostname
** On exit,
**	returns		0 if successful
**		        error code if not
*/
PRIVATE int nfs3processing
ARGS6 (
  char *,			name,
  HTParentAnchor *,		anchor,
  HTFormat,			format_out,
  HTStream *,			sink,
  CLIENT **,                    cl,
  int *,                        Version
)
{
      LOOKUP3res   *result_3;
      nfs_fh3      *Obj;
      fattr3       *Attr;
      char *filename = HTParse(name, "", PARSE_PATH + PARSE_PUNCTUATION);
      int          status = 0;
      char         *host = HTParse(name, "", PARSE_HOST);
      CLIENT       *clnt = *cl;

      nfs_error = NO_ERROR;
      

      /*  check whether we have to read the Root directory */
      if( (filename[0] == '\0')              /* read the root        */
	  || ((filename[1] == '\0') && (filename[0] == '/')) ) {
					     /* only '/' in the URL  */
	if(nfsCachePtr->PublicFH) {
	  nfs_fh3 obj = {0};
	  status = nfs3ReadDirectory(anchor, name, format_out, sink,
				   &obj, clnt);
	}
	else {
	  if(!nfs3StepLookUp(host, filename, &Obj, &Attr, clnt)) {
	    if(nfs_error != NFS_ERR_RPC)
	      return errCode;
	  }
	  else
	    return  nfs3ReadDirectory(anchor, name, format_out,
				      sink, Obj, clnt);
	}

	  
	/* if we get a STALE error, then the server does not yet support 
	 * public file handles. So, we do the `Mount` protocol          */ 
	if(nfs_error == NFS_ERR_STALE) {
	    nfsCachePtr->PublicFH = FALSE;
            StrAllocCopy(filename,
			 HTParse(name, "", PARSE_PATH + PARSE_PUNCTUATION) );
	    if(!nfs3StepLookUp(host, filename, &Obj, &Attr, clnt))
	       return errCode;
	    status = nfs3ReadDirectory(anchor, name, format_out,
				       sink, Obj, clnt);
	    return status;
	}
	  
	/*  check for NFS_RPC error. if so, it might be that the NFS 
	 *  Version for 'clnt' was created is down, and the other is up. 
	 *  So, in this case we destroy this client and create a new one 
	 *  and try lookup again                                         */
	else if(FirstTime && (nfs_error == NFS_ERR_RPC)) {
	    if(!ReplaceClient(host, name, cl, &Version))
	      return HT_NOT_LOADED;
	    else {
	      FirstTime = FALSE;
	      return NFS_STARTOVER;
	    }
	}
	else
	    return status;
      
      } /* end of "root" case */


      if ( filename[strlen(filename)-1] == '/' )
	filename[strlen(filename)-1] = '\0';

      status = 1;
      
      /*  lookup thru the directories  */
      if(nfsCachePtr->PublicFH) {
	 result_3 = nfs3LookUp(host, filename, clnt);
	 if(result_3 == NULL)
	   status = 0;
      }
      else 
	  status = nfs3StepLookUp(host, filename, &Obj, &Attr, clnt) ;

      if(!status) {
	   /* if we get a STALE error, then the server does not yet support 
	    * public file handles. So, we do the `Mount` protocol          */ 
	   if(nfs_error == NFS_ERR_STALE) {
             StrAllocCopy(filename,
			 HTParse(name, "", PARSE_PATH + PARSE_PUNCTUATION) );
	     nfsCachePtr->PublicFH = FALSE;
	     if(!nfs3StepLookUp(host, filename, &Obj, &Attr, clnt))
	       return errCode;
	   }
	  
	   /*  check for NFS_RPC error. if so, it might be that the NFS 
	    *  Version for 'clnt' was created is down, and the other is up. 
	    *  So, in this case we destroy this client and create a new one 
	    *  and try lookup again                                         */
	   else if(FirstTime && (nfs_error == NFS_ERR_RPC)) {
	     if(!ReplaceClient(host, name, cl, &Version))
	       return HT_NOT_LOADED;
	     else {
	       FirstTime = FALSE;
	       return NFS_STARTOVER;
	     }
	   }
	   else
	     return errCode;
      }
      else if(nfsCachePtr->PublicFH) {
	   Obj  = &result_3->LOOKUP3res_u.resok.object;
	   Attr = &result_3->LOOKUP3res_u.resok.obj_attributes.post_op_attr_u.attributes;
      }

	
      rootFlag = 0;
      
      /* check if we have to read the File or the Directory */
      if(Attr->type == NF3DIR) {
	/* read the directory */
	status = nfs3ReadDirectory(anchor, name, format_out, sink, Obj, clnt);
        if(status == NFS_MOUNTPOSSIBLE)
	  return nfs3ReadDirectory(anchor, name, format_out, sink, Obj, clnt);
	else
	  return status;
      }
      else 
	/* get the file */
	return nfs3ReadFile(filename, anchor, format_out, sink,
			    Obj, Attr, clnt);
}


/*	Perform NFS v2 processing
**	-------------------------
**
** On entry,
**	name		WWW address of a file: document, including hostname
** On exit,
**	returns		0 if successful
**		        error code if not
*/
PRIVATE int nfs2processing
ARGS6 (
  char *,			name,
  HTParentAnchor *,		anchor,
  HTFormat,			format_out,
  HTStream *,			sink,
  CLIENT **,                    cl,
  int *,                        Version
)
{
      diropres     *result_3;
      nfs_fh       *Obj;
      fattr        *Attr;
      char *filename = HTParse(name, "", PARSE_PATH + PARSE_PUNCTUATION);
      int          status = 0;
      char         *host = HTParse(name, "", PARSE_HOST);
      CLIENT       *clnt = *cl;

      nfs_error = NO_ERROR;
      

      /*  check whether we have to read the Root directory */
      if( (filename[0] == '\0')              /* read the root        */
	  || ((filename[1] == '\0') && (filename[0] == '/')) ) {
					     /* only '/' in the addr */
	if(nfsCachePtr->PublicFH) {
	  nfs_fh obj = {0};
	  status = nfs2ReadDirectory(anchor, name, format_out, sink,
				     &obj, clnt);
	}
	else {
	  if(!nfs2StepLookUp(host, filename, &Obj, &Attr, clnt)) {
	    if(nfs_error != NFS_ERR_RPC)
	      return errCode;
	  }
	  else
	    return nfs2ReadDirectory(anchor, name, format_out,
				     sink, Obj, clnt);
	}
	
	/* if we get a STALE error, then the server does not yet support 
	 * public file handles. So, we do the `Mount` protocol          */ 
	if(nfs_error == NFS_ERR_STALE) {
	    nfsCachePtr->PublicFH = FALSE;
            StrAllocCopy(filename,
			 HTParse(name, "", PARSE_PATH + PARSE_PUNCTUATION) );
            if(!nfs2StepLookUp(host, filename, &Obj, &Attr, clnt))
	       return errCode;
	    return nfs2ReadDirectory(anchor, name, format_out,
				     sink, Obj, clnt);
	}
 	  
	 /*  check for NFS_RPC error. if so, it might be that the NFS 
	  *  Version for 'clnt' was created is down, and the other is up. 
	  *  So, in this case we destroy this client and create a new one 
	  *  and try lookup again                                         */
	else if(FirstTime && (nfs_error == NFS_ERR_RPC)) {
	    if(!ReplaceClient(host, name, cl, &Version))
	      return HT_NOT_LOADED;
	    else {
	      FirstTime = FALSE;
	      return NFS_STARTOVER;
	    }
	}
	else
	    return status;
      
      } /* end of "root" case */

 
      if ( filename[strlen(filename)-1] == '/' )
	filename[strlen(filename)-1] = '\0';

      status = 1;
      
     /*  lookup thru the directories  */
      if(nfsCachePtr->PublicFH) {
	 result_3 = nfs2LookUp(host, filename, clnt);
	 if(result_3 == NULL)
	   status = 0;
      }
      else 
	  status = nfs2StepLookUp(host, filename, &Obj, &Attr, clnt) ;
	 
      if(!status) {
	   /* if we get a STALE error, then the server does not yet support 
	    * public file handles. So, we do the `Mount` protocol          */ 
	   if(nfs_error == NFS_ERR_STALE) {
	     StrAllocCopy(filename,
			  HTParse(name, "", PARSE_PATH + PARSE_PUNCTUATION) );
	     nfsCachePtr->PublicFH = FALSE;
	     if(!nfs2StepLookUp(host, filename, &Obj, &Attr, clnt))
	       return errCode;
	   }

	 
	   /*  check for NFS_RPC error. if so, it might be that that the NFS 
	    *  Version 'clnt' was created for is down, and the other is up. 
	    *  So, in this case we destroy this client and create a new one 
	    *  and try lookup again                                         */
	   else if(FirstTime && (nfs_error == NFS_ERR_RPC)) {
	     if(!ReplaceClient(host, name, cl, &Version))
	       return HT_NOT_LOADED;
	     else {
	       FirstTime = FALSE;
	       return NFS_STARTOVER;
	     }
	   }
	   else
	     return errCode;
      }
      else if(nfsCachePtr->PublicFH) {
	   Obj  = &result_3->diropres_u.diropres.file;
	   Attr = &result_3->diropres_u.diropres.attributes;
      }
 

      /* check if we have to read the File or the Directory */
      if(Attr->type == NF3DIR)
	/* read the directory */
	return nfs2ReadDirectory(anchor, name, format_out, sink,
				 Obj, clnt);
      else 
	/* get the file       */
	return nfs2ReadFile(filename, anchor, format_out, sink,
			    Obj, Attr, clnt);

}


/*	Init
**	----
**
**  initializes the environment for debugging
*/
PRIVATE void Init(void)
{
  char *host;
  char  name[256] ;
  char *nfs_filename = getenv("MOSAIC_NFSFILENAME");
  int   fStderr = 0;
  char *nfs_flag = getenv("MOSAIC_NFSDEBUGFLAG");

  strcpy(name, getenv("HOME"));
  strcat(name, "/.");
  
  if((nfs_flag != NULL) && (!strcasecmp("off", nfs_flag))) {
    if(NFSDEBUGFLAG)
      NFSDEBUGFLAG=0;
    return;
  }
  
  if(nfs_filename != NULL) {
   if(nfs_filename[0] == '\0') {
    if(!gethostname(host, MAXHOSTNAMELEN)) {
      if(strlen(host) > (240-strlen(name)))
	host[240-strlen(name)] = '\0';
      strcat(name, host);
    }
   }
   else if(!strcasecmp("stderr", nfs_filename)) 
    fStderr = 1;
   else
     strcat(name, nfs_filename);
  }
  
  strcat(name, "nfs-mosaic");
    
  if(fStderr) 
    debugFile = stderr;
  else 
    debugFile = fopen(name, "w");

  return;
}


/*	Retrieve File from Server
**	-------------------------
**
** On entry,
**	name		WWW address of a file: document, including hostname
** On exit,
**	returns		0 if successful
**			error code if not
*/
PUBLIC int HTNFSLoad
ARGS4 (
  char *,			name,
  HTParentAnchor *,		anchor,
  HTFormat,			format_out,
  HTStream *,			sink
)
{
  CLIENT     *clnt;
  char       *host;
  char       line[256];
  int        Version, status;
  static int fInit = 1;

  if(fInit) {
    fInit = 0;
    Init();
  }
  fprintf(debugFile, "\nHTNFSLoad : %s\n", name);
  
  numLinks   = 0;
  errCode    = 0;
  mountPoint = 1;
  FirstTime  = 1;

  host = HTParse(name, "", PARSE_HOST);
  if(!(strcmp(host, "localhost"))) {   /*  copy the hostname to `host` */
     host = (char*) malloc(MAXHOSTNAMELEN) ;
     gethostname(host, MAXHOSTNAMELEN);
  }

  /* no work done here! just for complying with the interface of the 
   * protocols                                                         */
  sprintf(line, "Looking up %s.", host);
  HTProgress(line);
  sprintf(line, "Making NFS connection to %s.", host);
  HTProgress(line);

  
  /* first search the Cache for hits */
  if(! SearchNfsCache(host, name, &clnt, &Version) )
      return HT_NOT_LOADED;

  /*  initialize the flag */
  rootFlag = 1;

nfsClientBegin:
  
  /*  get the required URL  */
  if(Version == NFS_V3) {
    status = nfs3processing(name, anchor, format_out, sink,
			    &clnt, &Version);
    if(status == NFS_STARTOVER)
      goto nfsClientBegin;
    else
      return status;
  }
  else {
    status = nfs2processing(name, anchor, format_out, sink,
			    &clnt, &Version);
    if(status == NFS_STARTOVER)
      goto nfsClientBegin;
    else
      return status;
  }
}
    
